<?php

namespace Drupal\field_redirect\Service;

use Drupal\Component\Render\MarkupInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\ContentEntityFormInterface;
use Drupal\Core\Entity\EditorialContentEntityBase;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileUrlGeneratorInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Link;
use Drupal\Core\Messenger\MessengerTrait;
use Drupal\Core\Render\Markup;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Url;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;
use Drupal\file\FileInterface;
use Drupal\node\NodeTypeInterface;
use Drupal\taxonomy\Entity\Vocabulary;
use Drupal\user\UserInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\HttpKernel\Exception\HttpException;

/**
 * Utility service class.
 */
class UtilityService {

  use StringTranslationTrait;
  use MessengerTrait;

  const MODULE_NAME = 'field_redirect';
  private const ENTITY_PAIR_SEPARATOR = ':';
  private const ENTITY_BUNDLE_WILDCARD = '*';
  private const PAIR_SEPARATOR = '|';
  private const FIELDS_SEPARATOR = ',';
  private const STATUS_SEPARATOR = '#';
  private const EOL = "\n";
  private const SUPPORTED_FIELD_TYPES = [
    'link',
    'file',
    'image',
  ];

  /**
   * The module settings.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $settings;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * The file URL generator.
   *
   * @var \Drupal\Core\File\FileUrlGeneratorInterface
   */
  protected $fileUrlGenerator;

  /**
   * The request service.
   *
   * @var \Symfony\Component\HttpFoundation\Request
   */
  protected $request;

  /**
   * The error message assoc.
   *
   * @var array
   */
  private $errors = [];

  /**
   * The line number.
   *
   * @var int
   */
  private $line_number = 0;

  /**
   * Service constructor.
   */
  public function __construct(
    ConfigFactoryInterface $config_factory,
    EntityTypeManagerInterface $entity_type_manager,
    EntityFieldManagerInterface $entity_field_manager,
    FileUrlGeneratorInterface $file_url_generator,
    RequestStack $request_stack,
  ) {
    $this->settings = $config_factory->get(self::MODULE_NAME . '.settings');
    $this->entityTypeManager = $entity_type_manager;
    $this->entityFieldManager = $entity_field_manager;
    $this->fileUrlGenerator = $file_url_generator;
    $this->request = $request_stack->getCurrentRequest();
  }

  /**
   * Convert the mappings array to a string.
   */
  public function toString(array $mappings_array = []): string {
    $mappings_string = '';
    foreach ($mappings_array as $entity_type__entity_bundle => $fields) {
      if (\strpos($entity_type__entity_bundle, 'user' . self::ENTITY_PAIR_SEPARATOR) === 0) {
        $entity_type__entity_bundle = 'user';
      }
      $mappings_string .= $this->desorptEntityAndFields('implode', [$entity_type__entity_bundle, $this->desorptFields('implode', $fields)]) . self::EOL;
    }
    return (string) $this->trim($mappings_string);
  }

  /**
   * Convert a mappings string to an array.
   */
  public function toArray(string $mappings_string = ''): array {
    $mappings_string = $this->trim($mappings_string);
    if ($mappings_string === '') {
      return [];
    }
    $mappings_array = [];
    if ($lines = $this->trim(\explode(self::EOL, $mappings_string))) {
      foreach ($lines as $entity_type_bundle__fields) {
        [$entity_type__entity_bundle, $fields] = $this->desorptEntityAndFields('explode', $entity_type_bundle__fields);
        $mappings_array[$entity_type__entity_bundle] = \array_map(fn ($field_machine_name__status) => $this->desorptFieldMachineNameAndStatus('explode', $field_machine_name__status), $this->desorptFields('explode', $fields));
      }
    }
    return $mappings_array;
  }

  /**
   * Smart trimmer.
   */
  public function trim(string|array $string): string|array {
    switch (\gettype($string)) {
      case 'array':
        foreach ($string as $k => $v) {
          $string[$k] = $this->trim($v);
        }
        return $string;

      case 'string':
        $string = \str_replace("\r\n", self::EOL, $string);
        $string = \str_replace("\r", self::EOL, $string);
        $string = \preg_replace('/\n+/', self::EOL, $string);
        return \mb_ereg_replace('^[[:space:]]*([\s\S]*?)[[:space:]]*$', '\1', $string);
    }
  }

  /**
   * Validate that the mappings is correct.
   */
  public function isValidMappings(FormStateInterface $form_state, array $element, string $mappings_string = ''): bool {
    $mappings_string = $this->trim($mappings_string);
    if ($mappings_string === '') {
      return TRUE;
    }
    if ($lines = $this->trim(\explode(self::EOL, $mappings_string))) {
      $this->errors = [];
      $this->line_number = 0;
      $duplicates_entity = [];
      foreach ($lines as $entity_type_bundle__fields) {
        if ($entity_type_bundle__fields === '') {
          continue;
        }
        ++$this->line_number;
        [$entity_type__entity_bundle, $fields] = $this->desorptEntityAndFields('explode', $entity_type_bundle__fields);
        [$entity_type, $entity_bundle] = $this->desorptTypeAndBundle('explode', $entity_type__entity_bundle);
        if ($entity_type === '' || ($entity_type !== 'user' && $entity_bundle === '') || $fields === '') {
          $this->mergeErrorMessage('This mapping format is incorrect.');
          continue;
        }
        $duplicates_entity[$entity_type__entity_bundle] ??= 0;
        ++$duplicates_entity[$entity_type__entity_bundle];
        $field_definitions = [];
        if ($entity_bundle === self::ENTITY_BUNDLE_WILDCARD && \in_array($entity_type, ['node', 'taxonomy_term', 'user'], TRUE)) {
          $field_definitions = $this->entityFieldManager->getFieldStorageDefinitions($entity_type);
        }
        else {
          switch ($entity_type) {
            case 'node':
              if ($this->entityTypeManager->getStorage('node_type')->load($entity_bundle) instanceof NodeTypeInterface) {
                $field_definitions = $this->entityFieldManager->getFieldDefinitions($entity_type, $entity_bundle);
              }
              break;

            case 'taxonomy_term':
              if ($this->entityTypeManager->getStorage('taxonomy_vocabulary')->load($entity_bundle) instanceof Vocabulary) {
                $field_definitions = $this->entityFieldManager->getFieldDefinitions($entity_type, $entity_bundle);
              }
              break;

            case 'user':
              // There doesn't seem to be a concept of bundles in the User entity.
              if ($entity_bundle === '') {
                $field_definitions = $this->entityFieldManager->getFieldStorageDefinitions($entity_type);
              }
              break;

            default:
              $this->mergeErrorMessage('Entity type "@entity_type" is not supported.', ['@entity_type' => $entity_type]);
              break;
          }
        }
        if (empty($field_definitions)) {
          $this->mergeErrorMessage('The bundle name "@entity_bundle" does not exist for entity type "@entity_type".', ['@entity_bundle' => $entity_bundle, '@entity_type' => $entity_type]);
        }
        $field_machine_names = $this->desorptFields('explode', $fields);
        $duplicates_field = [];
        $max_i = \count($field_machine_names) - 1;
        foreach ($field_machine_names as $current_i => $field_machine_name__status) {
          $duplicates_field[$field_machine_name__status] ??= 0;
          ++$duplicates_field[$field_machine_name__status];
          [$field_machine_name, $status] = $this->desorptFieldMachineNameAndStatus('explode', $field_machine_name__status);
          if (!empty($field_machine_name)) {
            if ($field_config = $field_definitions[$field_machine_name] ?? FALSE) {
              $field_type = '';
              switch (TRUE) {
                case ($field_config instanceof FieldConfig):
                  $field_type = $field_config->get('field_type');
                  break;

                case ($field_config instanceof FieldStorageConfig):
                  $field_type = $field_config->get('type');
                  break;
              }
              if (!\in_array($field_type, self::SUPPORTED_FIELD_TYPES, TRUE)) {
                $this->mergeErrorMessage('The field type for "@field_machine_name" must be link, file or image.', ['@field_machine_name' => $field_machine_name]);
              }
            }
            else {
              $this->mergeErrorMessage('The field for "@field_machine_name" does not exist for bundle "@entity_bundle" of entity type "@entity_type".', ['@field_machine_name' => $field_machine_name, '@entity_bundle' => $entity_bundle, '@entity_type' => $entity_type]);
            }
          }
          if (!empty($status)) {
            if (\strpos($status, '3') === 0) {
              if ($field_machine_name === '') {
                $this->mergeErrorMessage('The field should be specified if status "#3xx" is specified.');
              }
              elseif ($status !== '301' && $status !== '302' && $status !== '303' && $status !== '307') {
                $this->mergeErrorMessage('The specified status is invalid. Please specify "#301", "#302", "#303" or "#307".');
              }
            }
            if (\strpos($status, '4') === 0) {
              if ($field_machine_name !== '') {
                $this->mergeErrorMessage('The field cannot be specified if status "#4xx" is specified.');
              }
              elseif ($status !== '403' && $status !== '404') {
                $this->mergeErrorMessage('The specified status is invalid. Please specify "#403" or "#404".');
              }
              if ($max_i !== $current_i) {
                $this->mergeErrorMessage('Status is can be specified only one at the end.');
              }
            }
          }
        }
        foreach ($duplicates_field as $field => $count) {
          if ($count > 1) {
            if (\strpos($field, '#4') === 0) {
              $this->mergeErrorMessage('Multiple specification of status "#4xx" is not allowed.');
            }
            else {
              $this->mergeErrorMessage('Fields for "@field_machine_name" are specified is duplicate.', ['@field_machine_name' => $field_machine_name]);
            }
          }
        }
      }
    }
    foreach ($duplicates_entity as $entity_type_bundle => $count) {
      if ($count > 1) {
        $this->mergeErrorMessage('Set of the entity type and bundle is duplicated. (@entity_type_bundle)', ['@entity_type_bundle' => $entity_type_bundle], FALSE);
      }
    }
    if (!empty($this->errors)) {
      $form_state->setError($element, Markup::create(\implode('<br>', \array_keys($this->errors))));
      return FALSE;
    }
    return TRUE;
  }

  /**
   * Form alter.
   */
  public function formAlter(&$form, FormStateInterface $form_state, $form_id) {
    $form_object = $form_state->getFormObject();
    if ($form_object instanceof ContentEntityFormInterface) {
      $entity = $form_object->getEntity();
      if ($entity instanceof EntityInterface) {
        $redirect_fields = $this->getRedirectFields($entity->getEntityTypeId(), $entity->bundle());
        foreach ($redirect_fields as $field_machine_name__status) {
          [$field_machine_name, $status] = $field_machine_name__status;
          if ($widget = &$form[$field_machine_name]['widget'][0] ?? FALSE) {
            $message = $this->t('<p><strong>Field Redirect Module: If this field value is valid, redirect to the field value URL with redirect status "@status".</strong></p>', [
              '@status' => $this->getHttpStatusMessage($status),
            ]);
            if (isset($widget['uri'])) {
              $widget['uri']['#description'] = $message . $widget['uri']['#description'];
            }
            elseif (($widget['#type'] ?? '') === 'managed_file') {
              $widget['#description'] = $message . $widget['#description'];
            }
          }
        }
      }
    }
  }

  /**
   * Redirect to the specified destination or return the status.
   */
  public function doRedirect(string $mode, EntityInterface $entity): void {
    if ($entity = $this->checkEntityInstance($entity, TRUE)) {
      $target_entity_type = $entity->getEntityTypeId();
      $target_entity_bundle = $entity->bundle();
      $target_entity_id = $entity->id();

      // Check exists target field in entities bundle.
      $redirect_fields = $this->getRedirectFields($target_entity_type, $target_entity_bundle);

      $field_machine_name = '';
      $field_value = [];
      $status = '';
      foreach ($redirect_fields as $field_machine_name__status) {
        [$_field_machine_name, $status] = $field_machine_name__status;
        if ($entity->hasField($_field_machine_name) && ($_field_value = $entity->get($_field_machine_name)->getValue())) {
          $field_machine_name = $_field_machine_name;
          $field_value = $_field_value;
          break;
        }
      }
      if ($field_machine_name === '' && $status === '') {
        return;
      }

      // The "taxonomy_term" path seems to be a little different from the others.
      $target_entity_type_path
        = ($target_entity_type === 'taxonomy_term')
        ? 'taxonomy/term'
        : $target_entity_type;

      switch ($mode) {
        case 'after_save':
          // If exist "destination" in current url parameters then
          // this function process ending here, because to give
          // priority to "destination".
          if ($this->request->query->get('destination')) {
            return;
          }

          // Get label of entities bundle.
          $target_bundle_label = $this->getEntityTypeLabel($entity);

          // 4xx.
          if (\strpos($status, '4') === 0) {
            // Display a status message descriptioning this
            // behavior to avoid editor confusion.
            $this->messenger()->addStatus($this->t('Field Redirect Module: The original behavior was that if you accessed this %target_bundle_label view directly, then return status of the @status, but because I had just edited it, was redirected to the %target_bundle_label edit form.', [
              '%target_bundle_label' => $target_bundle_label,
              '@status' => $this->getHttpStatusMessage($status),
            ]));
          }

          // Redirect.
          else {
            // Display a status message descriptioning this
            // behavior to avoid editor confusion.
            $this->messenger()->addStatus($this->t('Field Redirect Module: The original behavior was that if you accessed this %target_bundle_label view directly, would be redirect to %a_tag (@status), but since it was just edited, was redirected to the %target_bundle_label edit form.', [
              '%target_bundle_label' => $target_bundle_label,
              '%a_tag' => $this->getUrl($field_value, TRUE),
              '@status' => $this->getHttpStatusMessage($status),
            ]));
          }

          // Return to the edit form.
          $redirect_path = '/' . $target_entity_type_path . '/' . $target_entity_id . '/edit';
          $redirect_url = Url::fromUserInput($redirect_path)->toString();
          $redirect_response = new RedirectResponse($redirect_url);
          $redirect_response->send();
          break;

        case 'view':
          // 4xx.
          if (\strpos($status, '4') === 0) {
            throw new HttpException($status);
          }

          // Redirect.
          else {
            // Prepare values.
            $redirect_url = $this->getUrl($field_value);

            // Avoid redirect loops.
            $site_url = $this->request->getSchemeAndHttpHost();
            $site_url__no_scheme = \preg_replace('/^[^:]+/', '', $site_url);
            $current_uri = $this->request->getRequestUri();
            $current_url__http = 'http' . $site_url__no_scheme . $current_uri;
            $current_url__https = 'https' . $site_url__no_scheme . $current_uri;
            $current_entity_uri = '/' . $target_entity_type_path . '/' . $target_entity_id;
            $current_entity_url__http = 'http' . $site_url__no_scheme . $current_entity_uri;
            $current_entity_url__https = 'https' . $site_url__no_scheme . $current_entity_uri;
            if (FALSE
              || $redirect_url === $current_uri
              || $redirect_url === $current_url__http
              || $redirect_url === $current_url__https
              || $redirect_url === $current_entity_uri
              || $redirect_url === $current_entity_url__http
              || $redirect_url === $current_entity_url__https
            ) {
              return;
            }

            // Do redirect to $redirect_url.
            (new RedirectResponse($redirect_url, $status ?: 307))
              ->send();
            exit;
          }
          break;
      }
    }
  }

  /**
   * Add error message without duplication.
   */
  private function mergeErrorMessage(string $message, array $args = [], bool $line_number = TRUE): void {
    if ($line_number) {
      $message = $this->t('Line @line_number:')->__toString() . ' ' . $message;
      $args += ['@line_number' => $this->line_number];
    }
    $error_key = (new TranslatableMarkup($message, $args))->__toString();
    $this->errors[$error_key] = 1;
  }

  /**
   * Explode and Implode the "entity type + entity bundle" & "fields" pair.
   */
  private function desorptEntityAndFields(string $mode, string|array $entity_type_bundle__fields): array|string {
    switch ($mode) {
      case 'explode':
        $pair = $this->trim(\explode(self::PAIR_SEPARATOR, $entity_type_bundle__fields, 2));
        $entity_type__entity_bundle = $pair[0] ?? '';
        $fields = $pair[1] ?? '';
        return [$entity_type__entity_bundle, $fields];

      case 'implode':
        $entity_type_bundle__fields = (array) $this->trim($entity_type_bundle__fields);
        if (empty($entity_type_bundle__fields)) {
          return '';
        }
        if (empty($entity_type_bundle__fields[1])) {
          return (string) $entity_type_bundle__fields[0];
        }
        return (string) \implode(self::PAIR_SEPARATOR, $entity_type_bundle__fields);
    }
  }

  /**
   * Explode and Implode the "entity type" & "entity bundle" pair.
   */
  private function desorptTypeAndBundle(string $mode, string|array $entity_type__entity_bundle): array|string {
    switch ($mode) {
      case 'explode':
        $pair = $this->trim(\explode(self::ENTITY_PAIR_SEPARATOR, $entity_type__entity_bundle, 2));
        $entity_type = $pair[0] ?? '';
        $entity_bundle = $pair[1] ?? '';
        return [$entity_type, $entity_bundle];

      case 'implode':
        $entity_type__entity_bundle = (array) $this->trim($entity_type__entity_bundle);
        if (empty($entity_type__entity_bundle)) {
          return '';
        }
        if (empty($entity_type__entity_bundle[1])) {
          return (string) $entity_type__entity_bundle[0];
        }
        return (string) \implode(self::ENTITY_PAIR_SEPARATOR, $entity_type__entity_bundle);
    }
  }

  /**
   * Explode and Implode the "fields".
   */
  private function desorptFields(string $mode, string|array $fields): array|string {
    switch ($mode) {
      case 'explode':
        return $this->trim(\explode(self::FIELDS_SEPARATOR, $fields));

      case 'implode':
        return (string) \implode(self::FIELDS_SEPARATOR, \array_map(fn ($field_machine_name__status) => $this->desorptFieldMachineNameAndStatus('implode', $field_machine_name__status), (array) $fields));
    }
  }

  /**
   * Explode and Implode the "field_machine_name" & "status" pair.
   */
  private function desorptFieldMachineNameAndStatus(string $mode, string|array $field_machine_name__status): array|string {
    switch ($mode) {
      case 'explode':
        $pair = $this->trim(\explode(self::STATUS_SEPARATOR, (string) $field_machine_name__status, 2));
        $field_machine_name = $pair[0] ?? '';
        $status = $pair[1] ?? '';
        return [$field_machine_name, $status];

      case 'implode':
        $field_machine_name__status = (array) $this->trim($field_machine_name__status);
        if (empty($field_machine_name__status)) {
          return '';
        }
        if (empty($field_machine_name__status[1])) {
          return (string) $field_machine_name__status[0];
        }
        return (string) \implode(self::STATUS_SEPARATOR, $field_machine_name__status);
    }
  }

  /**
   * Get redirect fields the specified bundle.
   */
  private function getRedirectFields(string $target_entity_type, string $target_entity_bundle): array {
    $mappings = $this->settings->get(self::MODULE_NAME . '.mappings');
    $wildcard_entity_type = $this->desorptTypeAndBundle('implode', [$target_entity_type, self::ENTITY_BUNDLE_WILDCARD]);
    $specified_entity_type = $this->desorptTypeAndBundle('implode', [$target_entity_type, $target_entity_bundle]);
    $redirect_fields = $mappings[$wildcard_entity_type] ?? [];
    $redirect_fields += $mappings[$specified_entity_type] ?? [];
    return $redirect_fields;
  }

  /**
   * Get label from specified entity.
   */
  private function getEntityTypeLabel(EntityInterface $entity): string {
    switch ($entity->getEntityTypeId()) {
      case 'node':
        $node_type = $this->entityTypeManager->getStorage('node_type')->load($entity->bundle());
        if ($node_type instanceof NodeTypeInterface) {
          return $node_type->get('name');
        }
        break;

      case 'taxonomy_term':
        $vocabulary = $this->entityTypeManager->getStorage('taxonomy_vocabulary')->load($entity->bundle());
        if ($vocabulary instanceof Vocabulary) {
          return $vocabulary->get('name');
        }
        break;

      case 'user':
        return $this->t('User');
    }
    return '[unknown]';
  }

  /**
   * Get html link tag markup.
   */
  private function getUrl(array $target_field_value, bool $need_link_tag = FALSE): MarkupInterface|string|null {
    // If exist value of URL.
    if ($uri = $target_field_value[0]['uri'] ?? FALSE) {
      // Generate of url from $uri.
      $url = Url::fromUri($uri, ['attributes' => ['target' => '_blank']]);

      return $need_link_tag
        ? Link::fromTextAndUrl($url->toString(), $url)->toString()
        : $url->toString();
    }

    // If exist value of target id of th file entity.
    elseif ($target_id = $target_field_value[0]['target_id'] ?? FALSE) {
      $file = $this->entityTypeManager->getStorage('file')->load($target_id);
      if ($file instanceof FileInterface) {
        if ($uri = $file->get('uri')->getString()) {
          // File URL.
          $url = $this->fileUrlGenerator->generateAbsoluteString($uri);

          // Link text.
          $text = $file->getFilename();
          if (($description = (string) $target_field_value[0]['description'] ?? '') !== '') {
            $text .= ' (' . $description . ')';
          }

          return $need_link_tag
            ? Markup::create('<a href="' . $url . '" target="_blank">' . $text . '</a>')
            : $url;
        }
      }
    }
    return NULL;
  }

  /**
   * Validate if an entity is allowed and return that entity.
   */
  private function checkEntityInstance(EntityInterface $entity, bool $check_status = FALSE): EditorialContentEntityBase|UserInterface|false {
    if ($entity instanceof EditorialContentEntityBase) {
      if ($check_status) {
        return $entity->isPublished() ? $entity : FALSE;
      }
      return $entity;
    }
    if ($entity instanceof UserInterface) {
      if ($check_status) {
        return $entity->isActive() ? $entity : FALSE;
      }
      return $entity;
    }
    return FALSE;
  }

  /**
   * Get the HTTP status message from the HTTP status code.
   */
  private function getHttpStatusMessage(string $status_code): string {
    switch ($status_code) {
      case '403':
        return '403 Forbidden';

      case '404':
        return '404 Not Found';

      case '301':
        return '301 Moved Permanently';

      case '302':
        return '302 Found';

      case '303':
        return '303 See Other';

      case '307':
        return '307 Temporary Redirect';
    }
    return '';
  }

}
